// MOD file player for SMSQ/E by Marcel Kilgus
//
// Based on HXCMOD by Jean Franois del Nero

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "qdos.h"
#include "qptr.h"
#include "hxcmod.h"

void *sss_check(void);
void sss_setm(void *sss, void **start, void **end);
void sss_addm(void *sss, void *next);
void sss_kill(void *sss);
size_t sss_sample(void *sss);


//void (*_consetup)() = consetup_title;
void (*_consetup)() = consetup_qpac;
char _prog_name[] = "MODPlay";
char _version[] = "v0.03";
char _copyright[] = "M. Kilgus & J.F. del Nero";
int _getcurspos(chanid_t chanid);


uint8_t *loadFile(const char *filename, size_t *size)
{
	FILE			*ff = 0;
	unsigned int	fsize;
	unsigned int	readsz;
	unsigned char	*data = 0;

	ff = fopen(filename, "rb");
	if (!ff) {
		printf("error opening\n", filename);
		return 0;
	}
	fseek(ff, 0, SEEK_END);
	fsize = ftell(ff);
	rewind(ff);

	data = malloc(fsize);
	if (!data) {
		printf("error reserving memory");
		return 0;
	}

	readsz = fread(data, 1, fsize, ff);
	fclose(ff);
	if (readsz != fsize) {
		printf("error reading\n");
		free(data);
		return 0;
	}
	*size = fsize;
	return data;
}


void printLineNonBlock(chanid_t ch, char *s)
{
		if (!sd_tab(ch, 0, 0) && !sd_clrrt(ch, 0)) {
				io_sstrg(ch, 0, s, strlen(s));
		}
}


void winMove(void)
{
	chanid_t ch = fgetchid(stdout);
	int pos = _getcurspos(ch);

	consetup_qpac_move(ch);
	sd_pixp(ch, -1, pos >> 16, pos & 0xFFFF);
}


int checkKey(chanid_t qstdin, int timeout)
{
	uint8_t key;

	if (!io_fbyte(qstdin, 0, (char*)&key)) {
		if (key == 27) return 2;
		if (key == 245) winMove();
	}
	return 0;
}


int main(int argc, char *argv[])
{
	const char	*filename;
	uint8_t 	*fileData;
	size_t		fileSize;
	modcontext	mod;
	void		*sss;
	uint8_t 	*start, *end;
	int	quit = 0;
	int	oldPos = -1;
	int	firstPass = 1;
	chanid_t	qstdin = fgetchid(stdin);
	chanid_t	qstdout = fgetchid(stdout);
	char		title[21];

	printf("%s by Marcel Kilgus & Jean Franois del Nero\n\n", _prog_name);

	if (argc != 2) {
		printf("Usage: modplay <mod-file>\n");
		return -1;
	}
	filename = argv[1];

	printf("Checking for sampled sound system... ");
	sss = sss_check();
	if (!sss) {
		printf("not found\n");
		return -1;
	}
	printf("ok\n");

	printf("Loading file %s... ", filename);
	fileData = loadFile(filename, &fileSize);
	if (!fileData) return -1;
	printf("ok\n");

	printf("Initialising MOD library... ");
	if (!hxcmod_init(&mod)) {
		printf("error\n");
		return -1;
	}
	printf("ok\n");

	printf("Configuring MOD library... ");
	if (!hxcmod_setcfg(&mod, 20000, 1, 0)) {
		printf("error\n");
		return -1;
	}
	printf("ok\n");

	printf("Initialising MOD file... ");
	if (!hxcmod_load(&mod, fileData, fileSize)) {
		printf("error\n");
		return -1;
	}
	printf("ok\n");

	// Fields are not NUL terminated :-(
	strncpy(title, (char*)mod.song.title, sizeof(title));
	title[20] = 0;
	printf("\nPlaying title %s\n\n", title);

	printf("Press ESC to quit, CTRL+F4 to move window\n\n\n");

	while (!quit) {
		mssize space, samples;

		// Force one key test without timeout
		quit = checkKey(qstdin, 0);
		if (quit) break;

		// Idle while queue is full enough ( >1s )
		while (!quit && sss_sample(sss) > 40000) {
			// 10ms timeout so we do not have a busy loop
			quit = checkKey(qstdin, 5);
		}
		if (quit) break;

		if (oldPos != mod.tablepos) {
			char tmp[16];
			sprintf(tmp, "Frame %i/%i", mod.tablepos + 1, mod.song.length);
			oldPos = mod.tablepos;

			// Use only non-blocking I/O here
			printLineNonBlock(qstdout, tmp);
		}
		sss_setm(sss, &start, &end);
		space = end - start;

		// Make sure any remaining rest in buffer is >= 20000 bytes
		if (space > 60000) space = 40000;
		samples = space >> 1;

		quit = !hxcmod_fillbuffer(&mod, start, &samples, 0);
		sss_addm(sss, start + samples * 2);
	}

	if (quit == 1) {
		printf("\n\nEnd of song...");
		while (sss_sample(sss) > 0);
	}
	sss_kill(sss);
	hxcmod_unload(&mod);

	printf("\n");
	if (quit) _endmsg = 0;
	return 0;
}
